theory bNat = Main :

datatype Pos =
    One
  | Bit Pos bool

datatype bNat =
    Zero  
  | PBin Pos


consts
  bSucc :: "bNat => bNat"
  pSucc :: "Pos => Pos"
  bPlus :: "bNat => bNat => bNat"
  bMult :: "bNat => bNat => bNat"
  pPlus :: "Pos => Pos => Pos"
  pMult :: "Pos => Pos => Pos"
  pos2nat :: "Pos => nat"
  nat2pos :: "nat => Pos"
  n2b :: "nat => bNat"
  b2n :: "bNat => nat"

(* We need these purely to be able to define a signature morphism from
 * Nat -> bNat (otherwise I don't know how to check type preservation
 * properly). *)

instance 
  bNat :: zero ..

defs 
  zero_def: "0 == Zero"

instance
  bNat :: plus ..

defs
  plus_def: "(op +) == bPlus"


primrec
  "bSucc Zero = PBin One"
  "bSucc (PBin x) = PBin (pSucc x)"

primrec
  "pSucc One = Bit One False"
  "pSucc (Bit x b)= (if b then (Bit (pSucc x)) (False)
                           else Bit x True)"

primrec
  "bPlus Zero x = x"
  "bPlus (PBin x) y = (case y of Zero => PBin x
                                 | PBin y => PBin (pPlus x y))"

primrec 
  "pPlus One y = pSucc y"
  "pPlus (Bit x b1) y = (case y of One => pSucc (Bit x b1)
                                     | (Bit z b2) => Bit (pPlus x (if (b1&b2) then pSucc z else z)) (b1 ~= b2))"

primrec
  "pMult One y = y"
  "pMult (Bit x b) y = (if b then (pPlus (Bit (pMult x y) False) y)
                                 else (Bit (pMult x y) False))"
(*
bin_mult (v BIT x) w =
                    (if x then (bin_add (NCons (bin_mult v w) False) w)
                          else (NCons (bin_mult v w) False))"
*)
primrec
  "bMult Zero y = Zero"
  "bMult (PBin x) y = (case y of Zero => Zero
                                 | (PBin y) => (PBin (pMult x y)))"

primrec
  "pos2nat One = 1"
  "pos2nat (Bit x b) = (if b then 1 else (0::nat)) + pos2nat x + pos2nat x"

primrec 
  "nat2pos 0 = One"
  "nat2pos (Suc n) = (if (n=1) then One else pSucc (nat2pos n))"

primrec
  "n2b 0 = Zero"
  "n2b (Suc n) = bSucc (n2b n)"

primrec
  "b2n Zero = 0"
  "b2n (PBin x) = (pos2nat x)"

end
