
(* --- Example for usage of a signature morphism to move theorems --- *)

(* Define a signature morphism from Nat to bNat  *)
val sg_n2b = SigMorph.make (theory "Nat") (theory "bNat")
               [("nat", SigMorph.Simple "bNat.bNat")]
	       [("0", "bNat.bNat.Zero"), ("Suc", "bNat.bSucc"), 
		("op +", "bNat.bPlus")];
    
(* Move across a few theorems *)

(* The first few are routine and need just induction and plus as proof ob'ns*)
val bNat_add_0_right = AbsTac.abs_translate sg_n2b add_0_right
    |> comp bNatInduct |> comp bPlus_Zero |> comp bPlus_bSucc;

val bNat_add_Suc_right= AbsTac.abs_translate sg_n2b add_Suc_right
    |> comp bNatInduct |> comp bPlus_Zero |> comp bPlus_bSucc;

val bNat_add_assoc= AbsTac.abs_translate sg_n2b add_assoc
    |> comp bNatInduct |> comp bPlus_Zero |> comp bPlus_bSucc;

val bNat_add_Suc_right = AbsTac.abs_translate sg_n2b add_Suc_right
    |> comp bNatInduct |> comp bPlus_Zero |> comp bPlus_bSucc;

val bNat_add_0_right = AbsTac.abs_translate sg_n2b add_0_right
    |> comp bNatInduct |> comp bPlus_Zero |> comp bPlus_bSucc;

(* add_commute needs some of the ones already moved *)
val bNat_add_commute= AbsTac.abs_translate sg_n2b add_commute
    |> comp bNatInduct |> comp bPlus_Zero |> comp bNat_add_0_right
    |> comp bPlus_bSucc |> comp bNat_add_Suc_right;

(* add_left_commute needs only two axioms *)
val abs_add_left_commute= AbsTac.abs_translate sg_n2b add_left_commute
    |> comp bNat_add_assoc |> comp bNat_add_commute;

(* Needs injectivity of bSucc, did not need that one before *)
val bNat_Suc_Suc_eq = AbsTac.abs_translate sg_n2b Suc_Suc_eq |> comp bSucc_inj;

val bNat_add_right_cancel= AbsTac.abs_translate sg_n2b add_right_cancel
    |> comp bNatInduct |> comp bNat_add_0_right |> comp bNat_add_Suc_right 
    |> comp bNat_Suc_Suc_eq;


(* ... you get the picture! Looks pretty schematic, and indeed: We can 
 * mechanise it using the dependency graph analysis-- example-depana.sml 
 * shows how. *)


