theory TailRec_nat 
imports Trafo newNat
begin

subsection {* Abstraction *}

consts
sum  :: "nuNat list \<Rightarrow> nuNat"
sum2 :: "(nuNat list \<times> nuNat) \<Rightarrow> nuNat"

recdef sum "measure length"
"sum l = (if l = [] then Zero else nuPlus (hd l) (sum (tl l)))"

recdef sum2 "measure (\<lambda> (l, x). length l)"
"sum2 (l, e) = (if l = [] then e else sum2 (tl l, nuPlus e (hd l)))"


lemma sum_Cons : " sum (x#xs) = nuPlus x (sum xs) " by simp

lemma sum2_Cons : " sum2 (x#xs, a) = sum2 (xs, nuPlus a x) " by simp


lemma nuPlus_1 : "nuPlus e (sum l) = sum2 (l, e)"
proof -
 have "\<forall>e. nuPlus e (sum l) = sum2 (l, e)" (is "?P l")
 proof (induct l)
   show "?P []" by simp 
  next
   fix x xs
   assume IH: "?P xs"
   show "?P (x#xs)"
   proof (rule allI, simp only: sum_Cons sum2_Cons)
    fix e 
    show "nuPlus e (nuPlus x (sum xs)) = sum2 (xs, nuPlus e x)" (is "?L = ?R")
    proof -
     have "?L = nuPlus (nuPlus e x) (sum xs)" by (rule nuPluss_assoc)
     also from IH[rule_format, of "nuPlus e x"] have "... = ?R" by assumption
     finally show ?thesis .
    qed
   qed
  qed
 thus ?thesis ..
qed



theorem equality : " sum l = sum2(l, Zero) "
proof -
 note nuPlus_1[of Zero l]
 thus ?thesis by simp
qed

(* Abstracting over the nuNat type *)
abstract_thm "equality" over "newNat.nuNat" as abs_eq 
expanding "TailRec_nat.nuPlus_1", "TailRec_nat.sum_Cons", "TailRec_nat.sum2_Cons"

(* Abstracting only over Zero and nuPlus constants *)
abstract_thm "equality" over_consts "newNat.nuNat.Zero" and "newNat.nuPlus" as abs_eq2 
expanding "newNat.nuPluss_assoc"


end






