
structure Parser_Ext =
struct

structure P = OuterParse and K = OuterKeyword;

fun save_mixfix thynm (s1, s2, mxfx) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => Syn_mng.add_mixfix (thynm ^ "." ^ s1, mxfx))
  in
   (s1, s2, mxfx)
  end 

fun save_mixfix2 thynm ((s1, s2, mxfx), s3) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => Syn_mng.add_mixfix (thynm ^ "." ^ s1, mxfx))
  in
   ((s1, s2, mxfx), s3)
  end 

fun save_mixfix3 thynm (l1, l2) =
 let fun f [] = ()
       | f ((opt, (x, s))::l) = (case opt of
                                NONE => (f l)
                               | SOME (s1, s2, mxfx) =>
                                  (let val _ = (case mxfx of 
                                                 NoSyn => ()
                                                | _ => Syn_mng.add_mixfix (thynm^"."^s1, mxfx))
                                   in
                                    f l
                                   end))
     val _ = f l2
 in
  (l1, l2)
 end


(* *** Types *** *)

fun add_typedecls' (a, args, mx) thy =
  let val thynm = Aux.name_of_thy thy
      val _ = (case mx of
                 NoSyn => ()
               | _ => Syn_mng.add_type_mixfix (thynm ^ "." ^ a, mx))
  in
   Theory.add_typedecls [(a, args, mx)] thy
  end   


val typedeclP =
  OuterSyntax.command "typedecl" "type declaration" K.thy_decl
    (P.type_args -- P.name -- P.opt_infix >> (fn ((args, a), mx) =>
      Toplevel.theory (add_typedecls' (a, args, mx))))



fun add_tyabbrs' abbrs thy =
  let val thynm = Aux.name_of_thy thy
      val _ = map (fn (a, args, T, mx) => (case mx of
                                            NoSyn => ()
                                           | _ => Syn_mng.add_type_mixfix (thynm ^ "." ^ a, mx))) abbrs
  in
   Theory.add_tyabbrs abbrs thy
  end

val typeabbrP =
  OuterSyntax.command "types" "declare type abbreviations" K.thy_decl
    (Scan.repeat1
      (P.type_args -- P.name -- (P.$$$ "=" |-- P.!!! (P.typ -- P.opt_infix')))
      >> (Toplevel.theory o add_tyabbrs' o
        map (fn ((args, a), (T, mx)) => (a, args, T, mx))))


(* *** Consts *** *)
fun add_consts' cl thy =
  let val thynm = Aux.name_of_thy thy
  in
   Theory.add_consts (map (save_mixfix thynm) cl) thy
  end

val constsP =
  OuterSyntax.command "consts" "declare constants" K.thy_decl
    (Scan.repeat1 P.const >> (Toplevel.theory o add_consts'));


(* *** Constdefs *** *)
fun add_constdefs' cl thy =
  let val thynm = Aux.name_of_thy thy
  in
   Constdefs.add_constdefs (save_mixfix3 thynm cl) thy
  end

val vars = P.and_list1 (Scan.repeat1 P.name -- Scan.option (P.$$$ "::" |-- P.typ));

val structs =
  Scan.optional ((P.$$$ "(" -- P.$$$ "structure") |-- P.!!! (vars --| P.$$$ ")")) [];

val constdecl =
  (P.name --| P.$$$ "where") >> (fn x => (x, NONE, Syntax.NoSyn)) ||
    P.name -- (P.$$$ "::" |-- P.!!! P.typ >> SOME) -- P.opt_mixfix' >> P.triple1 ||
    P.name -- (P.mixfix' >> pair NONE) >> P.triple2;

val constdef = Scan.option constdecl -- (P.opt_thm_name ":" -- P.prop);

val constdefsP =
  OuterSyntax.command "constdefs" "standard constant definitions" K.thy_decl
    (structs -- Scan.repeat1 constdef >> (Toplevel.theory o add_constdefs'));


(* *** Nonterminals *** *)

fun add_nonterminals' ntl thy =
  let val thynm = Aux.name_of_thy thy
      fun save_nt nt = 
         let val _ = Syn_mng.add_nonterm (thynm, nt)
         in 
          nt
         end
  in
   Theory.add_nonterminals (map save_nt ntl) thy
  end

val nontermP =
  OuterSyntax.command "nonterminals" "declare types treated as grammar nonterminal symbols"
    K.thy_decl (Scan.repeat1 P.name >> (Toplevel.theory o add_nonterminals'));


(* *** Syntax *** *)
fun add_modesyntax' ((s, b), sl) thy =
  let val thynm = Aux.name_of_thy thy
      fun save_syn (synnm, typestr, mxfx) = 
         let val typ = Sign.read_typ_syntax (thy, K NONE) typestr
         in
           Syn_mng.add_syn_decl (thynm, (s, b, synnm, typ, mxfx))
         end
      val _ = map save_syn sl
  in
   (uncurry Theory.add_modesyntax) ((s, b), sl) thy
  end

val mode_spec =
  (P.$$$ "output" >> K ("", false)) || P.name -- Scan.optional (P.$$$ "output" >> K false) true;

val opt_mode =
  Scan.optional (P.$$$ "(" |-- P.!!! (mode_spec --| P.$$$ ")")) Syntax.default_mode;

val syntaxP =
  OuterSyntax.command "syntax" "declare syntactic constants" K.thy_decl
    (opt_mode -- Scan.repeat1 P.const >> (Toplevel.theory o add_modesyntax')) 
                                           

(* *** Translations *** *)

fun add_trrules' trl thy =
 let val thyname = Aux.name_of_thy thy
     fun save_trrule tr =
         let val _ = Syn_mng.add_trrule (thyname, tr)
         in
          tr
         end
 in
  Theory.add_trrules (map save_trrule trl) thy
 end  

val trans_pat =
  Scan.optional (P.$$$ "(" |-- P.!!! (P.xname --| P.$$$ ")")) "logic" -- P.string;

fun trans_arrow toks =
  ((P.$$$ "\\<rightharpoonup>" || P.$$$ "=>") >> K Syntax.ParseRule ||
    (P.$$$ "\\<leftharpoondown>" || P.$$$ "<=") >> K Syntax.PrintRule ||
    (P.$$$ "\\<rightleftharpoons>" || P.$$$ "==") >> K Syntax.ParsePrintRule) toks;

val trans_line =
  trans_pat -- P.!!! (trans_arrow -- trans_pat)
    >> (fn (left, (arr, right)) => arr (left, right));

val translationsP =
  OuterSyntax.command "translations" "declare syntax translation rules" K.thy_decl
    (Scan.repeat1 trans_line >> (Toplevel.theory o add_trrules'));


(* *** Translation functions *** *)

fun parse_ast_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_parse_ast_trans (thyname, (b, s))
 in
  Sign.parse_ast_translation (b, s) thy
 end

fun parse_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_parse_trans (thyname, (b, s))
 in
  Sign.parse_translation (b, s) thy
 end

fun print_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_print_trans (thyname, (b, s))
 in
  Sign.print_translation (b, s) thy
 end

fun print_ast_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_print_ast_trans (thyname, (b, s))
 in
  Sign.print_ast_translation (b, s) thy
 end

fun token_translation' s thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_token_trans (thyname, s)
 in
  Sign.token_translation s thy
 end

val trfun = P.opt_keyword "advanced" -- P.text;

val parse_ast_translationP =
  OuterSyntax.command "parse_ast_translation" "install parse ast translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o parse_ast_translation'));

val parse_translationP =
  OuterSyntax.command "parse_translation" "install parse translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o parse_translation'));

val print_translationP =
  OuterSyntax.command "print_translation" "install print translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o print_translation'));

(* not yet 
val typed_print_translationP =
  OuterSyntax.command "typed_print_translation" "install typed print translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o Sign.typed_print_translation));
*)

val print_ast_translationP =
  OuterSyntax.command "print_ast_translation" "install print ast translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o print_ast_translation'));

val token_translationP =
  OuterSyntax.command "token_translation" "install token translation functions"
    (K.tag_ml K.thy_decl)
    (P.text >> (Toplevel.theory o token_translation'))


(* *** Theory *** *)

fun theory' (s, sl, l) t =
  let val t' = IsarThy.theory (s, sl, l) t
      val _ = Syn_mng.upd_rem_tabs s
  in
   t'
  end

val theoryP =
  OuterSyntax.command "theory" "begin theory" (K.tag_theory K.thy_begin)
    (ThyHeader.args >> (Toplevel.print oo theory'));


(* *** Records *** *)

fun add_record' a b c thy = (let val thynm = Aux.name_of_thy thy
                                 fun mk_nm s = (thynm ^ "." ^ (snd a) ^ "." ^ s)
                                 fun add_mxfxs [] = ()
                                   | add_mxfxs ((nm, tp, mxfx)::xs) =
                                      (let val fnm = mk_nm nm
                                           val _ = (case mxfx of
                                                     NoSyn => ()
                                                    | _ => Syn_mng.add_mixfix (fnm, mxfx))
                                       in
                                        add_mxfxs xs
                                       end)
                                 val _ = add_mxfxs c
                             in
                              RecordPackage.add_record a b c thy
                             end) 

val record_decl =
  P.type_args -- P.name --
    (P.$$$ "=" |-- Scan.option (P.typ --| P.$$$ "+") -- Scan.repeat1 P.const);

val recordP =
  OuterSyntax.command "record" "define extensible record" K.thy_decl
    (record_decl >> (fn (x, (y, z)) => Toplevel.theory (add_record' x y z)));



(* *** Datatypes *** *)

(* <thynm>.<typenm>.<consnm> *)

val datatype_decl =
  Scan.option (P.$$$ "(" |-- P.name --| P.$$$ ")") -- P.type_args -- P.name -- P.opt_infix --
    (P.$$$ "=" |-- P.enum1 "|" (P.name -- Scan.repeat P.typ -- P.opt_mixfix));

fun mk_datatype' args thy =
  let val thynm = Aux.name_of_thy thy
      fun add_mixfix [] = () 
        | add_mixfix ((vs, t, mx, cons)::xs) =
          (let val _ = (case mx of
                         NoSyn => ()
                       | _ => Syn_mng.add_type_mixfix (thynm ^ "." ^ t, mx))
          
               fun add_cons_mixfix [] = ()
                 | add_cons_mixfix ((nm, b, mx)::ys) =
                   (let val _ = (case mx of
                                   NoSyn => ()
                                 | _ => Syn_mng.add_mixfix (thynm ^ "." ^ t ^ "." ^ nm, mx))
                    in
                     add_cons_mixfix ys
                    end)
                val _ = add_cons_mixfix cons
           in 
             add_mixfix xs
           end)

    val names = map (fn ((((NONE, _), t), _), _) => t | ((((SOME t, _), _), _), _) => t) args;
    val specs = map (fn ((((_, vs), t), mx), cons) =>
      (vs, t, mx, map (fn ((x, y), z) => (x, y, z)) cons)) args;

    val _ = add_mixfix specs

  in #1(DatatypePackage.add_datatype false names specs thy) end;

val datatypeP =
  OuterSyntax.command "datatype" "define inductive datatypes" K.thy_decl
    (P.and_list1 datatype_decl >> (Toplevel.theory o mk_datatype'));





(* *** Redefine parsers *** *)

val _ = OuterSyntax.add_parsers [typeabbrP, constsP, nontermP, syntaxP, translationsP];
val _ = OuterSyntax.add_parsers [parse_ast_translationP, parse_translationP, 
                                 print_translationP, print_ast_translationP,
                                 token_translationP]
val _ = OuterSyntax.add_parsers [theoryP]
val _ = OuterSyntax.add_parsers [constdefsP]
val _ = OuterSyntax.add_parsers [recordP, datatypeP]



end